import argparse
import inspect
import os
import pkgutil
import sys
from pathlib import Path
from libcst.codemod import (
    CodemodCommand,
    VisitorBasedCodemodCommand,
    CodemodContext,
    gather_files,
    parallel_exec_transform_with_prettyprint,
)
from gitignore_parser import parse_gitignore
from pymender import commands

def run_command(Command: type[CodemodCommand], repo_root: str = ".") -> int:
    os.environ["LIBCST_PARSER_TYPE"] = "native"
    parser = argparse.ArgumentParser(
        description=Command.DESCRIPTION,
    )
    parser.add_argument(
        "path",
        metavar="PATH",
        nargs="+",
        help=(
            "Path to codemod. Can be a directory, file, or multiple of either. To "
            + 'instead read from stdin and write to stdout, use "-"'
        ),
    )
    parser.add_argument(
        "--gitignore",
        default=".gitignore",
        action="store",
        metavar="GITIGNORE",
        help=(
            "Path to a gitignore file. Files that match the gitignore will not be "
            + "codemodded."
        ),   
    )
    args = parser.parse_args()

    context = CodemodContext()
    command_instance = Command(context)
    files = gather_files(args.path)
    
    gitignore_path = Path(args.gitignore) if args.gitignore else Path(repo_root) / ".gitignore"
    if gitignore_path.exists() and (validator := parse_gitignore(gitignore_path)):
        files = [file for file in files if not validator(file)]
        
    try:
        result = parallel_exec_transform_with_prettyprint(
            command_instance, files, repo_root=repo_root,
        )
    except KeyboardInterrupt:
        print("Interrupted!", file=sys.stderr)
        return 2

    # Print a fancy summary at the end.
    print(
        f"Finished codemodding {result.successes + result.skips + result.failures} files!",
        file=sys.stderr,
    )
    print(f" - Transformed {result.successes} files successfully.", file=sys.stderr)
    print(f" - Skipped {result.skips} files.", file=sys.stderr)
    print(f" - Failed to codemod {result.failures} files.", file=sys.stderr)
    print(f" - {result.warnings} warnings were generated.", file=sys.stderr)
    if result.failures > 0:
        return 1
    return 0



def discover_commands() -> list[type[CodemodCommand]]:
    codemod_command_classes = set()
    for loader, name, _ in pkgutil.walk_packages(commands.__path__, commands.__name__ + '.'):
        module = loader.find_module(name).load_module(name)
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, CodemodCommand) and not inspect.isabstract(obj):
                codemod_command_classes.add(obj)
    codemod_command_classes -= {VisitorBasedCodemodCommand}
    return codemod_command_classes

def main() -> int:
    commands = discover_commands()
    command_lookup = {cmd.__name__: cmd for cmd in commands}
    
    parser = argparse.ArgumentParser(
        description="Mends Python code to be more idiomatic."
    )
    parser.add_argument(
        "command",
        metavar="COMMAND",
        choices=command_lookup.keys(),
        help="The codemod to run.",
    )
    args, _ = parser.parse_known_args()
    
    Command = command_lookup.get(args.command)
    if not Command:
        print(f"Unknown command {args.command}", file=sys.stderr)
        return 1
    return run_command(Command)
    


import logging
import sys

import libcst as cst
from libcst.codemod import VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor

logger = logging.getLogger(__name__)


def wrap_subscript_elements_with_annotated(
    param: cst.Param,
    elements: list[cst.BaseExpression | None],
    default: cst.BaseExpression | None,
) -> cst.Param:
    annotated = cst.Annotation(
        cst.Subscript(
            value=cst.Name("Annotated"),
            slice=tuple(
                [
                    cst.SubscriptElement(cst.Index(elem))
                    for elem in elements
                    if elem is not None
                ]
            ),
        )
    )
    return param.with_changes(
        annotation=annotated, equal=cst.MaybeSentinel.DEFAULT, default=default
    )


def adapt_param(param: cst.Param) -> cst.Param | None:
    match param.default:
        case cst.Call(func=cst.Name("Depends")):
            logger.info(f"Found Depends annotation")
            elements: list[cst.BaseExpression | None] = [
                param.annotation.annotation if param.annotation is not None else None,
                param.default,
            ]
            return wrap_subscript_elements_with_annotated(param, elements, None)
        case cst.Call(func=cst.Name("Body")) | cst.Call(func=cst.Name("Query")):
            logger.info(f"Found Body annotation")

            call = param.default

            # Find the default value (either first unnamed argument or default=)
            # breakpoint()

            # First unnamed argument, that is not Ellipsis
            first_unnamed_arg = next(
                (arg for arg in call.args if arg.keyword is None),
                None,
            )
            default_keyword = next(
                (
                    arg
                    for arg in call.args
                    if arg.keyword and arg.keyword.value == "default"
                ),
                None,
            )
            default = default_keyword or first_unnamed_arg

            if default is not None:
                call = call.with_changes(
                    args=[arg for arg in call.args if arg != default]
                )

            elements = [
                param.annotation.annotation if param.annotation is not None else None,
                call,
            ]
            return wrap_subscript_elements_with_annotated(
                param,
                elements,
                default.value
                if default and not isinstance(default.value, cst.Ellipsis)
                else None,
            )

        case _:
            return None


def sort_params_preserving_spacing(params: list[cst.Param]) -> list[cst.Param]:
    return [
        param.with_changes(comma=params[idx].comma)
        for idx, param in enumerate(
            sorted(params, key=lambda param: param.default is None, reverse=True)
        )
    ]


class FastAPIAnnotated(VisitorBasedCodemodCommand):
    DESCRIPTION = "Converts FastAPI annotations to use Annotated."

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        new_parameters: list[cst.Param] = []
        new_kwonlyparams: list[cst.Param] = []
        has_changes = False

        for param in node.params.params:
            adapted_param = adapt_param(param)
            if adapted_param is not None:
                new_parameters.append(adapted_param)
                has_changes = True
            else:
                new_parameters.append(param)

        for param in node.params.kwonly_params:
            adapted_param = adapt_param(param)
            if adapted_param is not None:
                new_kwonlyparams.append(adapted_param)
                has_changes = True
            else:
                new_kwonlyparams.append(param)

        if not has_changes:
            return

        # Sort the parameters to make sure default parameters are at the end
        # Preserve the original comma formats
        new_parameters = sort_params_preserving_spacing(new_parameters)
        new_kwonlyparams = sort_params_preserving_spacing(new_kwonlyparams)

        logger.debug(
            f"Converted {node.name.value} parameters: {new_parameters} * {new_kwonlyparams}"
        )

        self.context.scratch[node] = node.with_changes(
            params=node.params.with_changes(
                params=new_parameters, kwonly_params=new_kwonlyparams
            )
        )

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> cst.FunctionDef:
        new_node = self.context.scratch.get(original_node, updated_node)
        if original_node in self.context.scratch:
            AddImportsVisitor.add_needed_import(self.context, "typing", "Annotated")
        return new_node

